"use strict";
// Main module for passport-sqrl
Object.defineProperty(exports, "__esModule", { value: true });
const base64url_1 = require("base64url");
const crypto = require("crypto");
const ed25519 = require("ed25519");
const passport_strategy_1 = require("passport-strategy");
const urlLib = require("url");
// TODO: Support default implementation of encrypted nut and support TIFFlags.IPAddressesMatch
// TODO: Support disable, enable, remove
// TODO: Support suk client request and server validation
// TODO: Add urs= field
// TODO: Add UT for rekey request
// TODO: Test rekey against SQRL desktop implementation
/** Definitions for the Transaction Information Flag values specified in the SQRL specification. */
var TIFFlags;
(function (TIFFlags) {
    /** The web server found an identity association for the user based on the primary/current identity key. */
    TIFFlags[TIFFlags["CurrentIDMatch"] = 1] = "CurrentIDMatch";
    /** The web server found an identity association for the user based on the deprecated/previous identity key. */
    TIFFlags[TIFFlags["PreviousIDMatch"] = 2] = "PreviousIDMatch";
    /** The IP address seen at the server for this response is the same as the requester IP for the login page. */
    TIFFlags[TIFFlags["IPAddressesMatch"] = 4] = "IPAddressesMatch";
    /** The user's SQRL profile on the server has previously been marked disabled by the user. */
    TIFFlags[TIFFlags["IDDisabled"] = 8] = "IDDisabled";
    /** The client's request contained an unknown or unsupported verb. 0x40 CommandFailed will also be set in this case. */
    TIFFlags[TIFFlags["FunctionNotSupported"] = 16] = "FunctionNotSupported";
    /**
     * The server encountered an internal error and requests that the client reissue its request using the new nut
     * and query information in this response.
     */
    TIFFlags[TIFFlags["TransientError"] = 32] = "TransientError";
    /** The command failed. If 0x80 ClientFailure is not set, this indicates a non-retryable problem at the server. */
    TIFFlags[TIFFlags["CommandFailed"] = 64] = "CommandFailed";
    /** The command failure was because the client's request was malformed. */
    TIFFlags[TIFFlags["ClientFailure"] = 128] = "ClientFailure";
    /**
     * The SQRL ID specified in the client's request did not match the SQRL ID in ambient session
     * identity referred to by the client's cookie. The user needs to use the correct SQRL ID or
     * log out of the web site and log back in with a new identity.
     */
    TIFFlags[TIFFlags["BadIDAssociation"] = 256] = "BadIDAssociation";
})(TIFFlags = exports.TIFFlags || (exports.TIFFlags = {}));
/**
 * Returned from a completed ISQRLIdentityStorage function promise.
 */
class AuthCompletionInfo {
    constructor() {
        this.tifValues = 0;
    }
}
exports.AuthCompletionInfo = AuthCompletionInfo;
/**
 * Log levels for ILogger.
 */
var LogLevel;
(function (LogLevel) {
    LogLevel[LogLevel["Error"] = 0] = "Error";
    LogLevel[LogLevel["Warning"] = 1] = "Warning";
    LogLevel[LogLevel["Info"] = 2] = "Info";
    LogLevel[LogLevel["Debug"] = 3] = "Debug";
    LogLevel[LogLevel["Finest"] = 4] = "Finest";
})(LogLevel = exports.LogLevel || (exports.LogLevel = {}));
/**
 * ExpressJS middleware for the SQRL API.
 * Because SQRL does not use the HTTP Authenticate header in its data flow,
 * this handler is intended to be attached to a SQRL-specific route, e.g. '/sqrl',
 * that is not hooked into PassportJS. See the sample site in the passport-sqrl repo.
 */
class SQRLExpress {
    /**
     * Creates a new SQRL passport strategy instance.
     * @param identityStorage: Provides an identity storage implementation for calls from the SQRL layer.
     * @param log ILogger implementation for logging output. Allowed to be undefined/null.
     *   Errors are used for true errors in execution. Warnings are used for recoverable
     *   issues, e.g. errors caused by data from the network. Debug is for moderate
     *   frequency operational output. Finest is used for intensive traffic logging and
     *   detailed flow output.
     * @param config Configuration settings for this instance.
     */
    constructor(identityStorage, log, config) {
        /**
         * The Express middleware handler. Use like:
         *
         * let sqrlApi = new SQRLExpress(...);
         * app.post('/sqrl', sqrlApi.handleSqrlApi);
         */
        this.handleSqrlApi = (req, res) => {
            // Promisify to allow async coding style in authenticateAync and in unit tests.
            this.authenticateAsync(req)
                .then((authResult) => {
                this.log.debug('SQRL API call complete: ' +
                    `httpResponseCode: ${authResult.httpResponseCode}; ` +
                    `user: ${this.objToString(authResult.user)}; ` +
                    `encoded body: ${authResult.body}`);
                res.statusCode = authResult.httpResponseCode;
                res.send(authResult.body);
            })
                .catch(e => {
                let err = e; // Cast to commonly thrown error - if not correct, its httpStatusCode will be undefined.
                this.log.error(`Error thrown from SQRL API call: ${err} httpStatusCode=${err.httpStatusCode}`);
                // tslint:disable-next-line:no-bitwise
                let tif = TIFFlags.CommandFailed | TIFFlags.TransientError;
                // Per SQRL protocol, the name-value pairs below will be joined in the same order
                // with CR and LF characters, then base64url encoded.
                let nextNut = SqrlUrlFactory.nutToString(this.nutGenerator(req));
                let serverLines = [
                    'ver=1',
                    'nut=' + nextNut,
                    'tif=' + tif.toString(16),
                    'qry=' + this.config.urlPath + '?nut=' + nextNut,
                    // Use "ask" dialog on client to show error.
                    'ask=' + "Server error: " + err ? err.toString() : "<none>",
                ];
                let resp = serverLines.join("\r\n") + "\r\n"; // Last line must have CRLF as well.
                resp = base64url_1.default.encode(resp);
                res.statusCode = err.httpStatusCode || 500;
                res.send(resp);
            });
        };
        this.identityStorage = identityStorage;
        this.log = log;
        this.config = config;
        this.urlFactory = new SqrlUrlFactory(config.localDomainName, config.port, config.urlPath, config.domainExtension);
        if (!config.nutGenerator) {
            this.nutGenerator = this.generateRandomNut;
        }
        else {
            this.nutGenerator = config.nutGenerator;
        }
    }
    /**
     * Composes and returns a SQRL URL containing a unique "nut", plus the nut value for
     * registration for the external app login flow. The URL should be passed though a
     * QR-Code generator to produce the SQRL login QR for the client.
     */
    getSqrlUrl(req) {
        let nut = this.nutGenerator(req);
        let nutString = SqrlUrlFactory.nutToString(nut);
        return new UrlAndNut(this.urlFactory.create(nutString), nut, nutString);
    }
    /**
     * Promisified version of authenticate(). Not part of the public API.
     */
    async authenticateAsync(req) {
        let params;
        if (req.method === "POST") {
            params = req.body;
        }
        else {
            params = req.params; // Allow GET calls with URL params.
        }
        let clientRequestInfo = BodyParser.parseAndValidateRequestFields(params);
        if (clientRequestInfo.protocolVersion !== 1) {
            throw new ClientInputError(`This server only handles SQRL protocol revision 1`);
        }
        let nutInfoPromise = clientRequestInfo.nut ?
            this.identityStorage.getNutInfoAsync(clientRequestInfo.nut) : Promise.resolve(null);
        let nextNut = this.nutGenerator(req);
        let nextNutStr = SqrlUrlFactory.nutToString(nextNut);
        let nextUrl = this.config.urlPath + '?nut=' + nextNutStr;
        let urlAndNut = new UrlAndNut(nextUrl, nextNut, nextNutStr);
        let nutInfo = await nutInfoPromise;
        if (!nutInfo) {
            throw new ClientInputError('Client presented unknown nut value');
        }
        await this.identityStorage.nutIssuedToClientAsync(urlAndNut, nutInfo.originalLoginNut || nutInfo.nut);
        clientRequestInfo.nextNut = nextNutStr;
        clientRequestInfo.nextUrl = nextUrl;
        this.log.debug(`SQRL API call received to ${this.config.urlPath}: ` +
            `HTTP method ${req.method}. Parameter fields:${this.objToString(params)} . ` +
            'Decoded:' + this.objToString(clientRequestInfo));
        // The awaits here will throw any exceptions outward to the
        // authenticate() callback handler.
        let authCompletion;
        switch (clientRequestInfo.sqrlCommand) {
            case 'query':
                authCompletion = await this.identityStorage.queryAsync(clientRequestInfo, nutInfo);
                break;
            case 'ident':
                authCompletion = await this.identityStorage.identAsync(clientRequestInfo, nutInfo);
                break;
            case 'disable':
                authCompletion = await this.identityStorage.disableAsync(clientRequestInfo, nutInfo);
                break;
            case 'enable':
                authCompletion = await this.identityStorage.enableAsync(clientRequestInfo, nutInfo);
                break;
            case 'remove':
                authCompletion = await this.identityStorage.removeAsync(clientRequestInfo, nutInfo);
                break;
            default:
                throw new ClientInputError(`Unknown SQRL command ${clientRequestInfo.sqrlCommand}`);
        }
        this.log.debug(`Auth completion info: ${this.objToString(authCompletion)}`);
        return {
            user: authCompletion.user,
            body: this.authCompletionToResponseBody(clientRequestInfo, authCompletion),
            // Per the SQRL API for calls like query we must return a 200 even if
            // there is no login performed, as this is really an API endpoint with multiple
            // round-trips, and most of the error information is contained within the
            // SQRL response fields (e.g. TIF).
            httpResponseCode: 200,
        };
    }
    /** Default implementation of nut generation - creates a 128-bit random number. */
    generateRandomNut() {
        return crypto.randomBytes(16 /*128 bits*/);
    }
    authCompletionToResponseBody(clientRequestInfo, authInfo) {
        // Per SQRL protocol, the name-value pairs below will be joined in the same order
        // with CR and LF characters, then base64url encoded.
        let serverLines = [
            'ver=1',
            'nut=' + clientRequestInfo.nextNut,
            'tif=' + (authInfo.tifValues || 0).toString(16),
            'qry=' + this.config.urlPath + '?nut=' + clientRequestInfo.nextNut,
        ];
        if (clientRequestInfo.clientProvidedSession && clientRequestInfo.sqrlCommand !== 'query') {
            serverLines.push('url=' + this.config.clientLoginSuccessUrl);
        }
        if (clientRequestInfo.returnSessionUnlockKey && authInfo.sessionUnlockKey) {
            serverLines.push('suk=' + authInfo.sessionUnlockKey);
        }
        if (this.config.clientCancelAuthUrl) {
            serverLines.push('can=' + this.config.clientCancelAuthUrl);
        }
        let resp = serverLines.join("\r\n") + "\r\n"; // Last line must have CRLF as well.
        this.log.debug(`Response body pre-encoding: ${resp}`);
        resp = base64url_1.default.encode(resp);
        return resp;
    }
    objToString(o) {
        if (!o) {
            return "undefined";
        }
        let values = "";
        for (let propName in o) {
            values += ` ${propName}=${o[propName]}`;
        }
        return values;
    }
}
exports.SQRLExpress = SQRLExpress;
/**
 * SQRL Strategy for PassportJS.
 */
class SQRLStrategy extends passport_strategy_1.Strategy {
    constructor(log, config) {
        super();
        /**
         * The strategy name ('sqrl') to use when configuring a passport mapping
         * to an authentication route.
         */
        this.name = 'sqrl';
        this.log = log;
        this.config = config;
    }
}
exports.SQRLStrategy = SQRLStrategy;
/**
 * Encapsulates the parameters that are needed to resolve a SQRL API call.
 * Exported for unit testing but not intended to be used directly.
 */
class AuthenticateAsyncResult {
    constructor() {
        this.httpResponseCode = 500;
    }
}
exports.AuthenticateAsyncResult = AuthenticateAsyncResult;
/** A SQRL URL and its contained nut, broken out to separate fields for varying purposes. */
class UrlAndNut {
    constructor(url, nut, nutString) {
        this.url = url;
        this.nut = nut;
        this.nutString = nutString;
    }
}
exports.UrlAndNut = UrlAndNut;
/**
 * Parameters derived from the POST or GET parameters to the SQRL auth route.
 * This information gets passed to the query, ident, disable, enable, and remove
 * callbacks to the auth implementation. The public keys provided here have
 * already been validated against the private key and the nut and URL signatures.
 *
 * See https://www.grc.com/sqrl/protocol.htm particularly "How to form the POST verb's body."
 */
class ClientRequestInfo {
    constructor() {
        /** The client's SQRL protocol revision. */
        this.protocolVersion = 1;
        /**
         * Optional flag from the client ('sqrlonly' in its opt= option flag list, see
         * https://www.grc.com/sqrl/semantics.htm) in a non-query command, requesting
         * that the server disable other allowed authentication methods in favor of only SQRL.
         * The server should ignore this field value in query commands.
         */
        this.useSqrlIdentityOnly = false;
        /**
         * Optional flag from the client ('hardlock' in its opt= option flag list, see
         * https://www.grc.com/sqrl/semantics.htm) in a non-query command, requesting
         * that the server disable security question style alternate identity recovery methods.
         * The server should ignore this field value in query commands.
         */
        this.hardLockSqrlUse = false;
        /**
         * Optional flag from the client ('cps' in its opt= option flag list, see
         * https://www.grc.com/sqrl/semantics.htm) in an 'ident' command, requesting
         * that the server return in a url= response the logged-in URL that a client-side
         * plugin should redirect to after login is completed.
         */
        this.clientProvidedSession = false;
        /**
         * Optional flag from the client ('suk' in its opt= option flag list, see
         * https://www.grc.com/sqrl/semantics.htm) in a query command, requesting
         * that the server return its stored Session Unlock Key value to the client
         * so it can issue a re-key request.
         */
        this.returnSessionUnlockKey = false;
    }
    /** Provides a Buffer version of primaryIdentityPublicKey. */
    primaryIdentityPublicKeyBuf() {
        if (this.primaryIdentityPublicKey) {
            return Buffer.from(this.primaryIdentityPublicKey, 'base64');
        }
        return null;
    }
}
exports.ClientRequestInfo = ClientRequestInfo;
/** Provided to the SQRL strategy constructor to provide configuration information. */
class SQRLStrategyConfig {
    constructor() {
        /** Provides the domain name to use in generating SQRL URLs to send to clients. */
        this.localDomainName = 'please-configure-me.com';
    }
}
exports.SQRLStrategyConfig = SQRLStrategyConfig;
/** Data class containing information about a nut from identity storage. */
class NutInfo {
}
exports.NutInfo = NutInfo;
/** Error subclass that adds an HTTP status code. */
class ClientInputError extends Error {
    constructor(message, httpStatusCode = 400) {
        super(message);
        this.httpStatusCode = httpStatusCode;
    }
}
exports.ClientInputError = ClientInputError;
/** Parses and verifies the various parts of SQRL requests. Public for unit testing. */
class BodyParser {
    /** Reverses base64url encoding then parses the expected CRLF separated fields. */
    static parseBase64CRLFSeparatedFields(base64Props) {
        let preSplit = base64url_1.default.decode(base64Props);
        return BodyParser.parseCRLFSeparatedFields(preSplit);
    }
    static parseCRLFSeparatedFields(preSplitLines) {
        // The body is a base64url-encoded string that, when decoded, is a set of
        // name-value pairs separated by CRLF pairs (see https://www.grc.com/sqrl/protocol.htm).
        let lines = preSplitLines.split('\r\n');
        let props = {};
        lines.forEach(line => {
            line = line.trim();
            if (line.length === 0) {
                return;
            }
            // The name is considered everything up to the first = sign; the value everything after.
            let eqIndex = line.indexOf('=');
            if (eqIndex < 1) {
                throw new Error(`Failure parsing response line - no equal sign found in: ${line}`);
            }
            let name = line.substring(0, eqIndex);
            let val = line.substring(eqIndex + 1);
            props[name] = val;
        });
        return props;
    }
    /**
     * Parses the various POST body or GET URL parameter components passed from a SQRL client,
     * and validates that the client-provided signatures match.
     *
     * The fields can include (from the SQRL spec):
     * client: The 'client' string provided by the client, containing a
     *   base64url-encoded set of name-value pairs.
     * server: The 'server' string provided by the client, either a base64url form of a
     *   SQRL URL presented to the client, or the base64url-encoded body of a response
     *   from a previous API call.
     * idSignature The 'ids' string provided by the client, containing a base64url encoding of
     *   the 512-bit signature of the UTF-8 concatenation of the client and server strings, signed using the
     *   primary identity private key of the client for this server's domain.
     * prevIDSignatures Zero or more 'pids' string(s) provided by the client containing base64url encoded
     *   512-bit signatures of the UTF-8 concatenation of the client and server strings, signed using the
     *   corresponding private keys of the deprecated "previous IDs."
     * unlockRequestSignature An optional 'urs' string provided by the client, containing a base64url
     *   encoding of the 512-bit signature of the UTF-8 concatenation of the client and server strings,
     *   signed using the private Unlock Request Signing Key. The web server uses this The presence of this field and the corresponding
     */
    static parseAndValidateRequestFields(params) {
        if (!params) {
            throw new ClientInputError("Body is required");
        }
        if (!params.client) {
            throw new ClientInputError("Client field is required");
        }
        if (!params.server) {
            throw new ClientInputError("Server field is required");
        }
        let clientProps = BodyParser.parseBase64CRLFSeparatedFields(params.client);
        let requestInfo = {
            protocolVersion: Number(clientProps.ver),
            sqrlCommand: clientProps.cmd,
            primaryIdentityPublicKey: clientProps.idk,
            previousIdentityPublicKey: clientProps.pidk,
            serverUnlockPublicKey: clientProps.suk,
            serverVerifyUnlockPublicKey: clientProps.vuk,
            indexSecret: clientProps.ins,
            previousIndexSecret: clientProps.pins,
        };
        if (!requestInfo.primaryIdentityPublicKey) {
            throw new ClientInputError('Missing primary identity public key field in SQRL request');
        }
        if (!params.ids) {
            throw new ClientInputError('Missing ids= primary key signature field in SQRL request');
        }
        // Verify the client's primary and optional previous signatures versus the
        // client+server combined data.
        let clientServer = Buffer.from(params.client + params.server, 'utf8');
        let primaryKeySignature = Buffer.from(params.ids, 'base64');
        let primaryPublicKey = Buffer.from(requestInfo.primaryIdentityPublicKey, 'base64');
        let primaryOK = ed25519.Verify(clientServer, primaryKeySignature, primaryPublicKey);
        if (!primaryOK) {
            throw new ClientInputError('Primary public key did not verify correctly');
        }
        let previousOK = true;
        if (requestInfo.previousIdentityPublicKey) {
            if (!params.pids) {
                throw new ClientInputError('Missing pids= previous key signature field where the pidk= previous public key is specified');
            }
            let previousKeySignature = Buffer.from(params.pids, 'base64');
            let previousPublicKey = Buffer.from(requestInfo.previousIdentityPublicKey, 'base64');
            previousOK = ed25519.Verify(clientServer, previousKeySignature, previousPublicKey);
            if (!previousOK) {
                throw new ClientInputError('Previous public key did not verify correctly');
            }
        }
        // Decode the server= fields and verify a nut is present, and promote into the result.
        let serverDecoded = base64url_1.default.decode(params.server);
        if (serverDecoded.startsWith('sqrl')) {
            let qrCodeUrl = urlLib.parse(serverDecoded, /*parseQueryString:*/ true);
            requestInfo.nut = qrCodeUrl.query && typeof qrCodeUrl.query !== 'string' && qrCodeUrl.query.nut ? qrCodeUrl.query.nut.toString() : undefined;
        }
        else {
            let serverProps = BodyParser.parseBase64CRLFSeparatedFields(params.server);
            requestInfo.nut = serverProps.nut;
        }
        if (!requestInfo.nut) {
            throw new ClientInputError('server= info from client is not either a QR-code URL with nut= query param or a server response with nut= field');
        }
        if (clientProps.opt) {
            let optFields = clientProps.opt.toString().split('~');
            optFields.forEach(opt => {
                switch (opt) {
                    case 'sqrlonly':
                        requestInfo.useSqrlIdentityOnly = true;
                        break;
                    case 'hardlock':
                        requestInfo.hardLockSqrlUse = true;
                        break;
                    case 'cps':
                        requestInfo.clientProvidedSession = true;
                        break;
                    case 'suk':
                        requestInfo.returnSessionUnlockKey = true;
                        break;
                    default:
                        throw new ClientInputError(`Unknown SQRL client option ${opt}`);
                }
            });
        }
        return requestInfo;
    }
}
exports.BodyParser = BodyParser;
/**
 * Converts the bytes in the buffer to base64url and trims trailing '='
 * characters per the SQRL specification. Public for unit testing.
 */
function toSqrlBase64(buf) {
    return trimEqualsChars(base64url_1.default.encode(buf));
}
exports.toSqrlBase64 = toSqrlBase64;
/** Trims any tail '=' characters, returning the trimmed string. Public for unit testing. */
function trimEqualsChars(s) {
    // Avoid regular expressions - low performance.
    let len = s.length;
    if (!s || len === 0) {
        return s;
    }
    let i = len - 1;
    while (i >= 0) {
        if (s[i] === '=') {
            i--;
        }
        else {
            break;
        }
    }
    return s.substring(0, i + 1);
}
exports.trimEqualsChars = trimEqualsChars;
/**
 * Creates SQRL URLs. The static methods may be used directly, or else an instance
 * of this class may be instantiated with configuration information to reduce
 * the number of parameters a caller has to pass.
 */
class SqrlUrlFactory {
    /**
     * Creates a SQRL URL factory with static configuration information.
     * @param domain The site domain, e.g. "www.foo.com"
     * @param port Optional port.
     * @param pathString Optional path string, e.g. "path/to/sqrlLogin". May start with a forward slash.
     * @param domainExtension When positive, specifies the value to place into the x= query parameter
     *   that tells the client how many characters of the pathString to include in its server key hash.
     */
    constructor(domain, port, pathString, domainExtension) {
        this.domain = domain;
        this.port = port;
        this.pathString = pathString;
        this.domainExtension = domainExtension;
    }
    /**
     * Creates a SQRL URL from full "nut" metadata.
     * @param domain The site domain, e.g. "www.foo.com"
     * @param serverNut The opaque, unique server data generated for this URL, passed as the nut= query parameter.
     * @param pathString Optional path string, e.g. "path/to/sqrlLogin". May start with a forward slash.
     * @param domainExtension When positive, specifies the value to place into the x= query parameter that tells the client how many characters of the pathString to include in its server key hash.
     */
    static create(domain, serverNut, port, pathString, domainExtension) {
        let portPart = port ? `:${port}` : '';
        if (!pathString) {
            pathString = '';
        }
        if (pathString.length > 0 && pathString[0] !== '/') {
            pathString = '/' + pathString;
        }
        if (pathString.endsWith('?')) {
            pathString = pathString.substring(0, pathString.length - 1);
        }
        // domainExt includes the starting / of the path; start calculating after pathString has that prefix.
        let domainExt = '';
        if (pathString.length > 0 && domainExtension && domainExtension > 0) {
            domainExtension = Math.min(domainExtension, pathString.length);
            domainExt = `&x=${domainExtension}`;
        }
        let nut = SqrlUrlFactory.nutToString(serverNut);
        return `sqrl://${domain}${portPart}${pathString}?nut=${nut}${domainExt}`;
    }
    static nutToString(nut) {
        let nutStr;
        if (nut instanceof Buffer) {
            nutStr = toSqrlBase64(nut);
        }
        else {
            nutStr = nut;
        }
        return nutStr;
    }
    /**
     * Creates a SQRL URL from the provided unique server data.
     * @param serverNut The opaque, unique server data generated for this URL, passed as the nut= query parameter.
     * @param pathString Optional path string, e.g. "path/to/sqrlLogin". May start with a forward slash.
     *   Overrides any path specified in the constructor.
     * @param domainExtension When positive, specifies the value to place into the x= query parameter
     *   that tells the client how many characters of the pathString to include in its server key hash.
     *   Overrides any value specified in the constructor.
     */
    create(serverNut, pathString, domainExtension) {
        return SqrlUrlFactory.create(this.domain, serverNut, this.port, pathString || this.pathString, domainExtension || this.domainExtension);
    }
}
exports.SqrlUrlFactory = SqrlUrlFactory;
